{#- -*- mode: sgml; tab-width: 1; indent-tabs-mode: nil -*-
 #
 # SPDX-License-Identifier: MIT
-#}
{%- import('macros/printOdxCategory.xml.jinja2') as poc %}
{%- import('macros/printEcuSharedData.xml.jinja2') as pecusd -%}
{%- import('macros/printProtocol.xml.jinja2') as pprot %}
{%- import('macros/printFunctionalGroup.xml.jinja2') as pfuncgroup %}
{%- import('macros/printEcuSharedData.xml.jinja2') as pecusd %}
{%- import('macros/printBaseVariant.xml.jinja2') as pbv %}
{%- import('macros/printEcuVariant.xml.jinja2') as pecuv %}
{#- -#}

<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
<!-- Written using odxtools {{odxtools_version}} -->
<ODX MODEL-VERSION="2.2.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="odx.xsd">
 <DIAG-LAYER-CONTAINER {{- poc.printOdxCategoryAttribs(dlc) }}>
  {{- poc.printOdxCategorySubtags(dlc)|indent(3) }}
{%- if dlc.protocols %}
  <PROTOCOLS>
 {%- for dl in dlc.protocols %}
   {{pprot.printProtocol(dl)|indent(3)}}
 {%- endfor %}
  </PROTOCOLS>
{%- endif %}
{%- if dlc.functional_groups %}
  <FUNCTIONAL-GROUPS>
 {%- for dl in dlc.functional_groups %}
  {{pfuncgroup.printFunctionalGroup(dl)|indent(3)}}
 {%- endfor %}
  </FUNCTIONAL-GROUPS>
{%- endif %}
{%- if dlc.ecu_shared_datas %}
  <ECU-SHARED-DATAS>
 {%- for dl in dlc.ecu_shared_datas %}
  {{pecusd.printEcuSharedData(dl)|indent(3)}}
 {%- endfor %}
  </ECU-SHARED-DATAS>
{%- endif %}
{%- if dlc.base_variants %}
  <BASE-VARIANTS>
 {%- for dl in dlc.base_variants %}
  {{pbv.printBaseVariant(dl)|indent(3)}}
 {%- endfor %}
  </BASE-VARIANTS>
{%- endif %}
{%- if  dlc.ecu_variants %}
  <ECU-VARIANTS>
 {%- for dl in dlc.ecu_variants %}
  {{pecuv.printEcuVariant(dl)|indent(3)}}
 {%- endfor %}
  </ECU-VARIANTS>
{%- endif %}
 </DIAG-LAYER-CONTAINER>
</ODX>
