{#- -*- mode: sgml; tab-width: 1; indent-tabs-mode: nil -*-
 #
 # SPDX-License-Identifier: MIT
-#}

{%- import('macros/printElementId.xml.jinja2') as peid %}

{%- macro printStateTransition(state_transition) -%}
<STATE-TRANSITION {{-peid.printElementIdAttribs(state_transition)}}>
 {{ peid.printElementIdSubtags(state_transition)|indent(1) }}
 <SOURCE-SNREF SHORT-NAME="{{state_transition.source_snref}}" />
 <TARGET-SNREF SHORT-NAME="{{state_transition.target_snref}}" />
</STATE-TRANSITION>
{%- endmacro -%}
