{#- -*- mode: sgml; tab-width: 1; indent-tabs-mode: nil -*-
 #
 # SPDX-License-Identifier: MIT
-#}

{%- import('macros/printDescription.xml.jinja2') as pd %}

{%- macro printElementIdAttribs(obj) -%}
{#- #} {{- make_xml_attrib("ID", obj.odx_id.local_id) }}
{#- #} {{- make_xml_attrib("OID", getattr(obj, "oid", none)) -}}
{%- endmacro -%}

{%- macro printElementIdSubtags(obj) -%}
<SHORT-NAME>{{ obj.short_name }}</SHORT-NAME>
{%- if obj.long_name %}
<LONG-NAME>{{ obj.long_name|e }}</LONG-NAME>
{%- endif %}
{{- pd.printDescription(obj.description) }}
{%- endmacro -%}
