{#- -*- mode: sgml; tab-width: 1; indent-tabs-mode: nil -*-
 #
 # SPDX-License-Identifier: MIT
-#}

{%- import('macros/printElementId.xml.jinja2') as peid %}

{%- macro printState(state) -%}
<STATE {{-peid.printElementIdAttribs(state)}}>
 {{ peid.printElementIdSubtags(state)|indent(1) }}
</STATE>
{%- endmacro -%}
