{#- -*- mode: sgml; tab-width: 1; indent-tabs-mode: nil -*-
 #
 # SPDX-License-Identifier: MIT
-#}

{%- import('macros/printElementId.xml.jinja2') as peid %}
{%- import('macros/printParam.xml.jinja2') as pp %}

{%- macro printStructure(st) -%}
<STRUCTURE {{-peid.printElementIdAttribs(st)}}
           {{- make_bool_xml_attrib("IS-VISIBLE", st.is_visible_raw) }}{#- #}>
 {{ peid.printElementIdSubtags(st)|indent(1) }}
{%- if st.byte_size is not none %}
 <BYTE-SIZE>{{st.byte_size}}</BYTE-SIZE>
{%- endif %}
 <PARAMS>
{%- for param in st.parameters -%}
  {{ pp.printParam(param)|indent(2)}}
{%- endfor %}
 </PARAMS>
</STRUCTURE>
{%- endmacro -%}
