{#- -*- mode: sgml; tab-width: 1; indent-tabs-mode: nil -*-
 #
 # SPDX-License-Identifier: MIT
-#}

{%- import('macros/printElementId.xml.jinja2') as peid %}

{%- macro printEnvDataDesc(env_data_desc) %}
<ENV-DATA-DESC {{-peid.printElementIdAttribs(env_data_desc)}}>
  {{ peid.printElementIdSubtags(env_data_desc)|indent(1) }}
  <PARAM-SNREF SHORT-NAME="{{env_data_desc.param_snref}}"/>
  <ENV-DATA-REFS>
    {%- for env_data_ref in env_data_desc.env_data_refs %}
    <ENV-DATA-REF ID-REF="{{env_data_ref.ref_id}}"/>
    {%- endfor %}
  </ENV-DATA-REFS>
</ENV-DATA-DESC>
{%- endmacro -%}
